/* $Id$

This file is part of gldbg
Copyright © 2009-2010  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <signal.h>
#include <readline/readline.h>
#include <msp/core/except.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "commandinterpreter.h"
#include "enums.h"
#include "gldbg.h"
#include "tracer.h"

using namespace std;
using namespace Msp;

CommandInterpreter::CommandInterpreter(GlDbg &d):
	gldbg(d)
{
	register_command("help", this, &CommandInterpreter::cmd_help)
		.set_help("Provides help on commands",
			"help\n"
			"  Displays a list of commands\n\n"
			"help COMMAND\n"
			"  Gives detailed information on a command\n");
	register_command("exit", this, &CommandInterpreter::cmd_exit)
		.set_help("Ends the debugging session");
	commands["quit"] = new CommandAlias(commands["exit"]);

	register_command("run", this, &CommandInterpreter::cmd_run)
		.set_help("Starts the program");
	register_command("continue", this, &CommandInterpreter::cmd_continue)
		.set_help("Resumes program execution");
	register_command("kill", this, &CommandInterpreter::cmd_kill)
		.set_help("Terminates the program immediately");
	register_command("signal", this, &CommandInterpreter::cmd_signal)
		.set_help("Resumes execution with a signal",
			"signal NUM\n"
			"signal NAME\n"
			"  Sends the signal identified by NUM or NAME to the program and resumes\n"
			"  execution.  Currently recognized signal names are HUP, INT, TERM, SEGV\n"
			"  and TERM.\n");

}

void CommandInterpreter::execute(const string &cmd)
{
	unsigned space = cmd.find(' ');
	string name = cmd.substr(0, space);
	CommandMap::const_iterator i = commands.lower_bound(name);
	if(i==commands.end() || i->first.compare(0, name.size(), name))
		throw KeyError("Unknown command", name);
	if(i->first!=name)
	{
		CommandMap::const_iterator j = i;
		if((++j)!=commands.end() && !j->first.compare(0, name.size(), name))
			throw KeyError("Ambiguous command", name);
	}

	string args;
	if(space!=string::npos)
		args = cmd.substr(space+1);

	i->second->execute(args);
}

void CommandInterpreter::cmd_help(const string &args)
{
	if(args.empty())
	{
		for(CommandMap::const_iterator i=commands.begin(); i!=commands.end(); ++i)
			if(!dynamic_cast<const CommandAlias *>(i->second))
				IO::print("%-10s : %s\n", i->first, i->second->get_description());
	}
	else
	{
		CommandMap::const_iterator i = commands.find(args);
		if(i==commands.end())
			throw KeyError("Unknown command", args);

		const Command *cmd = i->second;
		while(const CommandAlias *alias = dynamic_cast<const CommandAlias *>(cmd))
			cmd = alias->get_target();

		IO::print("%s : %s\n", i->first, cmd->get_description());
		if(!cmd->get_help().empty())
			IO::print("\n%s", cmd->get_help());
	}
}

void CommandInterpreter::cmd_run(const string &)
{
	gldbg.launch();
}

void CommandInterpreter::cmd_continue(const string &)
{
	IO::print("Continuing.\n");
	gldbg.get_process().resume();
}

void CommandInterpreter::cmd_signal(const string &args)
{
	unsigned sig = 0;
	if(args=="HUP" || args=="SIGHUP")
		sig = SIGHUP;
	else if(args=="INT" || args=="SIGINT")
		sig = SIGINT;
	else if(args=="ILL" || args=="SIGILL")
		sig = SIGILL;
	else if(args=="SEGV" || args=="SIGSEGV")
		sig = SIGSEGV;
	else if(args=="TERM" || args=="SIGTERM")
		sig = SIGTERM;
	else if(isnumrc(args))
		sig = lexical_cast<unsigned>(args);
	else
		throw InvalidParameterValue("Invalid signal specification");
	gldbg.get_process().resume(sig);
}

void CommandInterpreter::cmd_kill(const string &)
{
	gldbg.get_process().kill();
}

void CommandInterpreter::cmd_exit(const string &)
{
	if(gldbg.get_process().get_state()!=Process::INACTIVE)
	{
		IO::print("Program is still running.  Kill it?\n");
		char *answer = readline("[y/n] ");
		if(answer[0]=='y')
		{
			gldbg.get_process().kill();
			gldbg.quit(true);
		}
		else
			IO::print("Not confirmed.\n");
	}
	else
		gldbg.quit(false);
}


void CommandInterpreter::Command::set_help(const string &d)
{
	description = d;
}

void CommandInterpreter::Command::set_help(const string &d, const string &h)
{
	description = d;
	help = h;
}


CommandInterpreter::CommandAlias::CommandAlias(Command *t):
	target(t)
{ }

void CommandInterpreter::CommandAlias::execute(const string &a)
{
	target->execute(a);
}
