/* $Id$

This file is part of gldbg
Copyright © 2009-2010  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <sys/uio.h>

#define INTERNAL __attribute__((visibility("internal")))

INTERNAL inline const char *get_lib_names(void)
{
	const char *env = getenv("GLWRAP_LIBS");
	if(env)
		return env;
	return "libGL.so";
}

INTERNAL inline void *glsym(const char *name)
{
	static void **gl_libs = NULL;
	unsigned i;

	if(!gl_libs)
	{
		char *lib_names = strdup(get_lib_names());
		unsigned n_libs = 1;
		unsigned j;

		for(i=0; lib_names[i]; ++i)
			if(lib_names[i]==':')
				++n_libs;

		gl_libs = (void **)malloc((n_libs+1)*sizeof(void *));
		i = 0;
		n_libs = 0;
		for(j=0;; ++j)
		{
			if(lib_names[j]==':' || lib_names[j]==0)
			{
				int at_end = (lib_names[j]==0);
				lib_names[j] = 0;

				gl_libs[n_libs] = dlopen(lib_names+i, RTLD_NOW);
				if(!gl_libs[n_libs])
				{
					fprintf(stderr, "Could not open %s: %s\n", lib_names+i, dlerror());
					abort();
				}

				i = j+1;
				++n_libs;

				if(at_end)
					break;
			}
		}

		gl_libs[n_libs] = 0;
		free(lib_names);
	}

	for(i=0; gl_libs[i]; ++i)
	{
		void *sym = dlsym(gl_libs[i], name);
		if(sym)
			return sym;
	}

	return NULL;
}

INTERNAL char *buffer = 0;
INTERNAL char *write_pos;
INTERNAL struct iovec *iovecs = 0;
INTERNAL struct iovec *cur_vec;
INTERNAL unsigned length;

INTERNAL inline void next_vec(void)
{
	if(write_pos!=cur_vec->iov_base)
	{
		cur_vec->iov_len = write_pos-(char *)cur_vec->iov_base;
		length += cur_vec->iov_len;
		++cur_vec;
		cur_vec->iov_base = write_pos;
	}
}

INTERNAL inline void write_bytes(const char *ptr, unsigned size)
{
	unsigned i;
	for(i=0; i<size; ++i)
		*write_pos++ = *ptr++;
}

INTERNAL inline void write_char(char v)
{
	*write_pos++ = v;
}

INTERNAL inline void write_short(short v)
{
	write_bytes((char *)&v, sizeof(short));
}

INTERNAL inline void write_int(int v)
{
	write_bytes((char *)&v, sizeof(int));
}

INTERNAL inline void write_long(long v)
{
	write_bytes((char *)&v, sizeof(long));
}

INTERNAL inline void write_long_long(long long v)
{
	write_bytes((char *)&v, sizeof(long long));
}

INTERNAL inline void write_float(float v)
{
	write_bytes((char *)&v, sizeof(float));
}

INTERNAL inline void write_double(double v)
{
	write_bytes((char *)&v, sizeof(double));
}

INTERNAL inline void write_pointer(const void *p)
{
	write_bytes((char *)&p, sizeof(void *));
}

INTERNAL inline void write_data(const void *data, unsigned size)
{
	if(data)
	{
		write_int(size);
		next_vec();
		cur_vec->iov_base = (void *)data;
		cur_vec->iov_len = size;
		length += size;
		++cur_vec;
		cur_vec->iov_base = write_pos;
	}
	else
		write_int(0);
}

INTERNAL inline void write_string(const char *s)
{
	write_data(s, strlen(s)+1);
}

INTERNAL inline void write_string_array(const char **sa, unsigned size)
{
	unsigned i;
	size /= sizeof(const char *);
	write_int(size);
	for(i=0; i<size; ++i)
		write_string(sa[i]);
}

INTERNAL inline void begin_packet(int func)
{
	if(!buffer)
		buffer = (char *)malloc(1024);
	if(!iovecs)
		iovecs = (struct iovec *)malloc(16*sizeof(struct iovec));

	write_pos = buffer;
	cur_vec = iovecs;
	cur_vec->iov_base = write_pos;
	length = 0;

	write_int(0);
	write_short(func);
}

INTERNAL inline int get_out_fd(void)
{
	static int fd = -1;

	if(fd<0)
	{
		const char *var = getenv("GLWRAP_FD");
		if(var)
			fd = strtol(var, NULL, 0);
		else
		{
			var = getenv("GLWRAP_FILE");
			if(var)
			{
				fd = open(var, O_WRONLY|O_CREAT, 0644);
				if(fd==-1)
				{
					fprintf(stderr, "Couldn't open dumpfile %s for output: %s", var, strerror(errno));
					abort();
				}
			}
			else
				fd = 2;
		}
	}

	return fd;
}

INTERNAL inline void send_partial_packet(void)
{
	next_vec();
	write_pos = buffer;
	write_int(length|0x80000000);
	writev(get_out_fd(), iovecs, cur_vec-iovecs);

	write_pos = buffer;
	cur_vec = iovecs;
	cur_vec->iov_base = write_pos;
	length = 0;

	write_int(0);
}

INTERNAL inline void send_packet(void)
{
	next_vec();
	write_pos = buffer;
	write_int(length);
	writev(get_out_fd(), iovecs, cur_vec-iovecs);
}
